#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "type.h"
#include "random.h"

double Decode(ChromType *genes, GlobalAppType *a);
void MakeFA(GlobalAppType *a, float *FA, int nSteps, IPTR pj);
double BinToDec(ChromType *gene, int size);
void InitCalcs(GlobalAppType *a);
void Smooth(float *FA, int size);
void PrintPH(FILE *fp, IPTR pj, int nSteps, Population *p);
void PrintFA(FILE *fp, IPTR pj, int nSteps, Population *p);

/* should be in utils */
void DecToBin(int val, int size, ChromType *bits);
void CopyBits(ChromType *bits, ChromType *gene, int size);
/* should be in utils */

GlobalAppType app;

void CTSMutate(IPTR pj, float pMut)
{
  int i, start, end, val, nSteps, 
    limit, maxVal, minVal, flag, flag2; 
  GlobalAppType *a;
  int nBits;
  ChromType *bits;

  if(!Flip(pMut)) return;
  
  a = &app;
  nSteps = pj->chromLen/a->nBitsPerGene + 1; 
  start = Rnd(0, (nSteps - 1)/2);
  limit = Rnd(0, nSteps/2);
  if(start + limit >= nSteps){
    limit = nSteps - start;
  }

  nBits = a->nBitsPerGene;

  bits = (ChromType *) malloc (sizeof(ChromType) * nBits);
  maxVal = pow(2.0, (double) nBits) - 1;
  minVal = 0;
  flag = Flip(0.5);
  flag2 = Flip(0.5);
  val = BinToDec(&(pj->chrom[start * nBits]), nBits); 
    /*    Rnd(0, maxVal); */
  for(i = start; i < start + limit; i++){
    if(flag2) {
      val = (flag ? val + 1: val - 1);
      if(val > maxVal) val = maxVal;
      if(val < minVal) val = minVal;
    }
    DecToBin(val, nBits, bits);
    CopyBits(bits, &(pj->chrom[i * nBits]), nBits);
  }
  free(bits);
}

void DecToBin(int val, int size, ChromType *bits)
{
  int i;
  for(i = 0; i < size; i++){
    bits[i] = val%2;
    val = val/2;
  }
}

void CopyBits(ChromType *bits, ChromType *gene, int size)
{
  int i;
  for(i = 0; i < size; i++){
    gene[i] = bits[i];
  }
}

double Eval(IPTR pj)
{
  int i, nSteps;
  double IAE, IAEnew;
  double penalty1, penalty2;
  double t1, t2, t3;
  double CCl, CNa, CH, D1, D2, sq;
  float *FA;
  GlobalAppType *a;

  a = &app;
  nSteps = pj->chromLen/a->nBitsPerGene + 1;
  FA = (float *) malloc (sizeof(float) * nSteps);

  /* Initialize */
  CCl = a->CClInit;
  CNa = a->CNaInit;
  CH  = a->CHInit;
  FA[0] = a->Fa0;
  pj->backup[0] = (float) (-1.0 * (log10(CH)));
  /*  printf("%.3f ", pj->backup[0]);*/
  /* end initialize */
  MakeFA(a, FA, nSteps, pj);
  /*printf("%.3f ", FA[0]); */

  for(i = 1; i < nSteps; i++){
    D1 = (FA[i - 1] * a->Ca - (a->Fb + FA[i - 1] * CCl))/a->Vol;
    D2 = (a->Fb     * a->Cb - (a->Fb + FA[i - 1] * CNa))/a->Vol;
    CCl = CCl + a->deltaT * D1;
    CNa = CNa + a->deltaT * D2;
    sq = CCl - CNa;
    sq = sq * sq;
    CH = (CCl - CNa + sqrt(sq  + 4.0 * a->Keq))/2.0;
    pj->backup[i] = (float) (-1.0 *  (log10(CH)));
  }

  IAE = 0.0;
  IAEnew = 0.0;
  penalty1 = 0.0;
  penalty2 = 0.0;
  for(i = 1; i < nSteps; i++){
    t1 = pj->backup[i] - a->SP;
    t2 = pj->backup[i] - pj->backup[i - 1];
    t3 = FA[i] - FA[i - 1];
    IAE += a->deltaT * fabs(t1);
    /*    IAEnew += (nSteps - i) * 
	  a->deltaT * fabs((double) (pj->backup[i] - a->SP));*/
    IAEnew += (nSteps - i) *
      a->deltaT * fabs(t1);
    penalty1 += i *
      a->deltaT * fabs(t2);
    penalty2 += i * i * 
      a->deltaT * fabs(t3);
  }

  free(FA);
  pj->dx = IAE;
  pj->dy = penalty1;
  pj->cx = penalty2;
  return (IAEnew + penalty2); /*  + penalty1/3.0 + penalty2/3.0);  + penalty1/100.0);*/
}

void MakeFA(GlobalAppType *a, float *FA, int nSteps, IPTR pj)
{
  int i, j;
  
  FA[0] = a->Fa0;
  for(j = 1, i = 0; j < nSteps && i < pj->chromLen; i += a->nBitsPerGene, j++){
    FA[j] = FA[j - 1] + Decode(&(pj->chrom[i]), a);
    if(FA[j] > a->upperLimit){
      FA[j] = a->upperLimit;
    }
    if(FA[j] < a->lowerLimit){
      FA[j] = a->lowerLimit;
    }
    
  }
  /*   Smooth(FA, nSteps);*/
}

void Smooth(float *FA, int nSteps)
{
  int i;
  float *copy;
  copy = (float *) malloc (sizeof(float) * nSteps);

  for(i = 0; i < nSteps; i++){
    copy[i] = FA[i];
  }
  for(i = 1; i < nSteps - 1; i++){
    copy[i] = (FA[i-1] + FA[i]) / 2.0;
  }
  for(i = 1; i < nSteps; i++){
    FA[i] = copy[i];
  }
  free(copy);
}




double Decode(ChromType *genes, GlobalAppType *a)
{
  double sum, val;

  sum = 0.0;
  val = BinToDec(genes, a->nBitsPerGene);
  sum = a->lowerStep + (a->precision * val);
  return sum;

}

double BinToDec(ChromType *gene, int size)
{
  int i;
  double sum;
  sum = 0.0;
  for(i = 0; i < size; i++){
    if(gene[i] == 1){
      sum += pow(2.0, (double) i);
      /*      printf("%i: sum: %f\n", i, sum); */
    }
  }
  return sum;
}


void AppInitChrom(IPTR pj)
{
  return;
}

void AppInit(char *appInfile)
{
 FILE *fp;
 
 if( (fp = fopen(appInfile, "r")) == NULL){ 
   printf("error in opening file %s \n", appInfile);
   exit(1);
  }

 fscanf(fp, "%f", &(app.Ca));
 printf("Ca: %f\n", app.Ca);
 fscanf(fp, "%f", &(app.Cb));
 printf("Cb: %f\n", app.Cb);
 fscanf(fp, "%f", &(app.Fb0));
 printf("Fb0: %f\n", app.Fb0);
 fscanf(fp, "%f", &(app.Fa0));
 printf("Fa0: %f\n", app.Fa0);
 fscanf(fp, "%f", &(app.Fb));
 printf("Fb: %f\n", app.Fb);
 fscanf(fp, "%f", &(app.Vol));
 printf("Vol: %f\n", app.Vol);
 fscanf(fp, "%lf", &(app.Keq));
 printf("Keq: %.20f\n", app.Keq);
 fscanf(fp, "%f", &(app.SP));
 printf("SP: %f\n", app.SP);
 fscanf(fp, "%f", &(app.deltaT));
 printf("deltaT: %f\n", app.deltaT);

 fscanf(fp, "%f", &(app.upperLimit));
 printf("upperLimit: %f\n", app.upperLimit);
 fscanf(fp, "%f", &(app.lowerLimit));
 printf("lowerLimit: %f\n", app.lowerLimit);
 fscanf(fp, "%f", &(app.upperStep));
 printf("upperStep: %f\n", app.upperStep);
 fscanf(fp, "%f", &(app.lowerStep));
 printf("lowerStep: %f\n", app.lowerStep);

 fscanf(fp, "%i", &(app.nBitsPerGene));
 printf("nBitsPerGene: %i\n", app.nBitsPerGene);

 fclose(fp);

 InitCalcs(&app);
}

void InitCalcs(GlobalAppType *a)
{
  double sq;

  a->CClInit = a->Fa0 * a->Ca/(a->Fa0 + a->Fb0);
  a->CNaInit = a->Fb0 * a->Cb/(a->Fa0 + a->Fb0);
  sq = a->CClInit - a->CNaInit;
  sq = sq * sq;
  a->CHInit = (a->CClInit - a->CNaInit + sqrt(sq + 4.0 * a->Keq))/2.0;

  printf("initial Ph = %f\n", -1.0 * log10((double) a->CHInit));

  a->precision = 
    (a->upperStep - a->lowerStep)/(pow(2.0, (double) a->nBitsPerGene) - 1); 
}


void PhenoPrint(FILE *fp, IPTR apop, Population *p)
{
  int  nSteps;
  GlobalAppType *a;
  IPTR pj;

  a = &app;
  pj = &(apop[p->maxi]);
  nSteps = pj->chromLen/a->nBitsPerGene + 1;

  PrintPH(fp, pj, nSteps, p);

  fclose(fp);

  if((fp = fopen(p->errorFile, "a")) == NULL){
    fprintf(stderr, "PrintFA: Cannot open %s\nExiting...", p->errorFile);
    exit(1);
  }

  PrintFA(fp, pj, nSteps, p);
}

void PrintPH(FILE *fp, IPTR pj, int nSteps, Population *p)
{
  int i;

  fprintf(fp, "%4i ", p->generation);
  for(i = 0; i < nSteps; i++){ 
    fprintf(fp, "%i %.3f ", i, pj->backup[i]); 
  }  
  fprintf(fp, "\n"); 

}

void PrintFA(FILE *fp, IPTR pj, int nSteps, Population *p)
{
  int i;
  float *FA;

  FA = (float *) malloc (sizeof(float) * nSteps);

  MakeFA(&app, FA, nSteps, pj);
  
  fprintf(fp, "%4i ", p->generation);

  for(i = 0; i < nSteps; i++){ 
    fprintf(fp, "%i %.10f ", i, FA[i]); 
  } 
  fprintf(fp, "\n"); 
  free(FA);
}



void InitPhenoPrint(IPTR pj, char *fname, Population *p)
{
  int i, nSteps;
  GlobalAppType *a;
  FILE *fp;
  
 if( (fp = fopen(fname, "w")) == NULL){ 
   printf("InitPhenoPrint: error in opening file %s \n", fname);
   exit(1);
  }

  a = &app;
  nSteps = p->chromLength/a->nBitsPerGene + 1;
  fprintf(fp, "%i\nPH: inum ", 2 * nSteps + 1);
  for(i = 0; i < nSteps; i++){
    fprintf(fp, "x y ");
  }
  fprintf(fp, "\n");
  fclose(fp);

 if( (fp = fopen(p->errorFile, "w")) == NULL){ 
   printf("InitPhenoPrint: error in opening file %s \n", p->errorFile);
   exit(1);
  }

  fprintf(fp, "%i\nFA: inum ", 2 * nSteps + 1);
  for(i = 0; i < nSteps; i++){
    fprintf(fp, "x y ");
  }
  fprintf(fp, "\n");
  fclose(fp);


}
